# -*- coding: utf-8 -*-
import logging

from housekeeper.pipelines.mip.build import build_analysis
from housekeeper.pipelines.mip.parse import prepare_inputs, parse_references
from housekeeper.pipelines.mip.core import build_assets, link_asset
from .prepare import prepare_run

log = logging.getLogger(__name__)


def parse(config_data, reference_data, force=False):
    """Parse MIP analysis output."""
    # 1. parse segments of input data
    segments = prepare_inputs(config_data)
    # 2. post-process the output a bit
    prepare_run(segments, force=force)
    # 3. build the records
    mip_version = segments['family'].get('MIPVersion', 'v2.x')
    new_objs = build_analysis(segments, version=mip_version)
    # 4. parse references
    new_refs = parse_references(reference_data, segments=segments)
    # 5. build assets from references + link to new records
    new_assets = build_assets(new_refs)
    for new_asset, sample in new_assets:
        link_asset(new_objs['run'], new_asset, sample=sample)
    return new_objs



# -*- coding: utf-8 -*-
from datetime import datetime
import json

import alchy
from path import Path
from flask_login import UserMixin
from sqlalchemy import Column, ForeignKey, orm, types, UniqueConstraint

from housekeeper.constants import PIPELINES, ARCHIVE_TYPES, EXTRA_STATUSES

SEQUENCERS = ('hiseqx', 'hiseq2500', 'hiseqHO', 'miseq', 'nextseq')
SAMPLE_CATEGORIES = ('wgs', 'wes', 'tga', 'wgs-ext', 'wes-ext', 'tga-ext')


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


class JsonModel(alchy.ModelBase):

    def to_json(self, pretty=False):
        """Serialize Model to JSON."""
        kwargs = dict(indent=4, sort_keys=True) if pretty else dict()
        return json.dumps(self.to_dict(), default=json_serial, **kwargs)


Model = alchy.make_declarative_base(Base=JsonModel)


class SampleRunLink(Model):

    """Link between UniqueSample and AnalysisRun."""

    __table_args__ = (
        UniqueConstraint('sample_id', 'run_id', name='_sample_run_uc'),
    )

    id = Column(types.Integer, primary_key=True)
    sample_id = Column(ForeignKey('sample.id'), nullable=False)
    run_id = Column(ForeignKey('analysis_run.id'), nullable=False)


class Sample(Model):

    """Store unique reference to a sample."""

    id = Column(types.Integer, primary_key=True)
    lims_id = Column(types.String(32), nullable=False, unique=True)
    case_id = Column(ForeignKey('case.id'))
    priority = Column(types.Boolean, default=False)
    _flowcells = Column(types.Text)
    category = Column(types.Enum(*SAMPLE_CATEGORIES))

    created_at = Column(types.DateTime, default=datetime.now)
    received_at = Column(types.DateTime)
    sequenced_at = Column(types.DateTime)
    confirmed_at = Column(types.DateTime)

    assets = orm.relationship('Asset', backref='sample', cascade='all,delete')
    runs = orm.relationship('AnalysisRun', secondary='sample_run_link',
                            back_populates='samples')

    @property
    def flowcells(self):
        """Return a list of flowcells."""
        flowcell_list = self._flowcells.split(',') if self._flowcells else []
        return flowcell_list

    @flowcells.setter
    def flowcells(self, flowcell_list):
        self._flowcells = ','.join(flowcell_list) if flowcell_list else None


class Case(Model):

    """Store a permanent reference to a case."""

    id = Column(types.Integer, primary_key=True)
    name = Column(types.String(128), unique=True, nullable=False)
    customer = Column(types.String(32), nullable=False)
    family_id = Column(types.String(128), nullable=False)
    created_at = Column(types.DateTime, default=datetime.now)
    is_onhold = Column(types.Boolean, default=False)
    comment = Column(types.Text)
    is_started = Column(types.Boolean, default=False)
    is_manual = Column(types.Boolean)
    rerun_requested = Column(types.Boolean, default=False)

    runs = orm.relationship('AnalysisRun', cascade='all,delete',
                            backref='case', order_by='-AnalysisRun.id')
    samples = orm.relationship('Sample', backref='case', cascade='all,delete')

    @property
    def has_priority(self):
        """Check if related sample is priority."""
        return any(sample.priority for sample in self.samples)

    @property
    def current(self):
        """Return the current (active) run."""
        if self.runs:
            return self.runs[0]
        else:
            return None


class AnalysisRun(Model):

    """Store information about a specific analysis run."""

    __table_args__ = (
        UniqueConstraint('case_id', 'analyzed_at', name='_uc_case_analyzed'),
    )

    id = Column(types.Integer, primary_key=True)
    created_at = Column(types.DateTime, default=datetime.now)
    pipeline = Column(types.Enum(*PIPELINES))
    pipeline_version = Column(types.String(32))
    reference_genome = Column(types.String(64))
    # keep track of a date if the cases is requested to be rerun
    requested_at = Column(types.DateTime)
    analyzed_at = Column(types.DateTime)
    compiled_at = Column(types.DateTime)
    delivered_at = Column(types.DateTime)
    answeredout_at = Column(types.DateTime)
    archived_at = Column(types.DateTime)
    cleanedup_at = Column(types.DateTime)
    will_cleanup_at = Column(types.DateTime)

    case_id = Column(ForeignKey(Case.id), nullable=False)
    assets = orm.relationship('Asset', cascade='all,delete', backref='run')
    samples = orm.relationship('Sample', secondary='sample_run_link',
                               back_populates='runs')
    extra = orm.relationship('ExtraRunData', backref='run',
                             cascade='all,delete', uselist=False)

    @property
    def cleanup_in(self):
        """Return number of days until archive happens."""
        time_diff = self.will_cleanup_at - datetime.today()
        return time_diff.days

    @property
    def sample_map(self):
        """Return dict of samples."""
        return {sample.lims_id: sample for sample in self.samples}

    @property
    def analysis_date(self):
        """Date of the analysis run."""
        return self.analyzed_at.date()

    def suggested_delivery(self):
        """Suggest date for delivery based on upload dates."""
        uploads = []
        for column in EXTRA_STATUSES:
            column_key = "{}_date".format(column)
            upload_date = getattr(self.extra, column_key)
            if upload_date:
                uploads.append(upload_date)
        latest_date = max(uploads)
        return latest_date.date() if latest_date else None

    def to_dict(self, skip_samples=False):
        """Also include samples in the dict serialization."""
        analysis_dict = super(AnalysisRun, self).to_dict()
        analysis_dict['name'] = self.case.name
        if not skip_samples:
            sample_dicts = [sample.to_dict() for sample in self.samples]
            analysis_dict['samples'] = sample_dicts
        return analysis_dict


class Asset(Model):

    """Asset/file belonging to an analysis."""

    id = Column(types.Integer, primary_key=True)
    original_path = Column(types.Text)
    path = Column(types.String(256), nullable=False, unique=True)
    category = Column(types.String(32))
    archive_type = Column(types.Enum(*ARCHIVE_TYPES))
    is_local = Column(types.Boolean, default=True)
    checksum = Column(types.String(128))

    run_id = Column(ForeignKey(AnalysisRun.id), nullable=False)
    sample_id = Column(ForeignKey('sample.id'))

    def basename(self):
        return Path(self.original_path).basename()


class User(Model, UserMixin):
    id = Column(types.Integer, primary_key=True)
    google_id = Column(types.String(128), unique=True)
    email = Column(types.String(128), unique=True)
    name = Column(types.String(128))
    avatar = Column(types.Text)

    @property
    def first_name(self):
        """First part of name."""
        return self.name.split(' ')[0]


class ExtraRunData(Model):

    """Store custom extra fields related to a run."""

    id = Column(types.Integer, primary_key=True)
    run_id = Column(ForeignKey('analysis_run.id'), nullable=False, unique=True)

    # custom dates
    coverage_date = Column(types.DateTime)
    frequency_date = Column(types.DateTime)
    genotype_date = Column(types.DateTime)
    visualizer_date = Column(types.DateTime)
    rawdata_date = Column(types.DateTime)
    qc_date = Column(types.DateTime)



# -*- coding: utf-8 -*-
import codecs
import os.path
import subprocess
import shutil
from tempfile import NamedTemporaryFile, mkdtemp

from ped_parser import FamilyParser


class SinglePedigreeError(Exception):
    pass


class MadelineIncompatibleError(Exception):
    pass


def run(ped_stream, sample_map, exe):
    """Run Madeline and capture the output."""
    ped = FamilyParser(ped_stream, family_type='alt', cmms_check=False)
    external_individuals = external_ped(ped.individuals, sample_map)
    family_id = ped.families.keys()[0]
    ped.families[family_id].individuals = external_individuals

    madeline_exe = os.path.abspath(exe)
    output_dir = mkdtemp()
    output_prefix = "{}/madeline".format(output_dir)
    out_path = "{}.xml".format(output_prefix)

    if not any([ind.has_parents for ind in ped.individuals.values()]):
        raise MadelineIncompatibleError("Madeline needs connected individuals")

    madeline_ped_lines = list(ped.to_madeline())
    if len(madeline_ped_lines) == 2:
        raise SinglePedigreeError("can't generate pedigree with single"
                                  "individual")

    # write the input to a temp file
    with NamedTemporaryFile('r+w') as in_file:
        madeline_content = '\n'.join(madeline_ped_lines)
        in_file.write(madeline_content)
        in_file.flush()

        subprocess.call([madeline_exe, '--color', '--outputprefix',
                         output_prefix, in_file.name])

    with codecs.open(out_path, 'r') as output:
        svg_content = output.read()

    # strip away the script tag
    script_tag = ('<script type="text/javascript" xlink:href='
                  '"javascript/madeline.js"></script>')
    svg_content.replace(script_tag, '')

    shutil.rmtree(output_dir)
    return svg_content


def external_ped(ped_individuals, sample_map):
    """Convert and fill out 'ped' with external ids."""
    external_individuals = {}
    for individual_id, individual in ped_individuals.items():
        external_id = sample_map.get(individual_id)
        individual.individual_id = external_id
        individual.mother = sample_map.get(individual.mother, '0')
        individual.father = sample_map.get(individual.father, '0')
        # all individuals in the pedigree have been sequenced
        individual.consultand = 'Yes'
        external_individuals[external_id] = individual

    return external_individuals



# -*- coding: utf-8 -*-
import ruamel.yaml

from housekeeper.exc import MalformattedPedigreeError


def write_meta(meta_output, outdata_dir):
    """Write meta files."""
    meta_path = "{}/meta.yaml".format(outdata_dir)
    with open(meta_path, 'w') as out_handle:
        out_handle.write(meta_output)


def build_meta(case_name, family_data, qc_ped, version=None, strict=True):
    """Build metadata information content.

    Args:
        run (AnalysisRun): analysis run record
        qc_ped (path): path to qc pedigree file

    Returns:
        str: YAML-formatted string of the data
    """
    with open(qc_ped, 'r') as stream:
        qcped_data = ruamel.yaml.safe_load(stream)
    sample_map = sampleid_map(qcped_data, strict=strict)
    metadata = {
        'name': case_name,
        'pipeline': 'mip',
        'pipeline_version': version or family_data['MIPVersion'],
        'analyzed_at': family_data['AnalysisDate'],
        'samples': sample_map,
    }
    return ruamel.yaml.dump(metadata, Dumper=ruamel.yaml.RoundTripDumper)


def sampleid_map(qc_ped, strict=True):
    """Take out information about internal/external sample names."""
    fam_key = list(qc_ped.keys())[0]
    samples = {}
    for sample in qc_ped[fam_key].values():
        sample_id = sample.get('Individual ID', sample.get('SampleID'))
        if sample_id is None:
            raise MalformattedPedigreeError(fam_key)
        sample_names = sample.get('display_name')
        if sample_names is None and not strict:
            sample_names = sample.get('Display_name')
        if isinstance(sample_names, list):
            samples[sample_id] = sample_names[0]
        else:
            message = "{}: 'display_name'".format(fam_key)
            raise MalformattedPedigreeError(message)

    return samples



# -*- coding: utf-8 -*-
import logging

from path import Path

from housekeeper.store import AnalysisRun, api, ExtraRunData
from housekeeper.store.utils import get_rundir
from housekeeper.exc import AnalysisConflictError

log = logging.getLogger(__name__)


def check_existing(name, run_obj):
    """Check if the analysis is already added."""
    same_date = AnalysisRun.analyzed_at == run_obj.analyzed_at
    old_run = api.runs(name).filter(same_date).first()
    return old_run


def analysis(manager, root_path, case_obj, new_run):
    """Store an analysis run with files to the backend."""
    run_root = get_rundir(root_path, case_obj.name, new_run)
    if run_root.isdir():
        raise AnalysisConflictError("'{}' analysis run output exists"
                                    .format(run_root))

    for asset in new_run.assets:
        filename = asset.basename()
        log.info("adding asset: %s", filename)
        new_path = run_root.joinpath(filename)
        asset.path = new_path

    log.debug("commit new analysis to database")
    new_run.extra = ExtraRunData()
    case_obj.runs.append(new_run)
    manager.add_commit(case_obj)
    # reset the rerun request if any
    case_obj.rerun_requested = False

    # perform linking to the new structure
    run_root.makedirs_p()
    try:
        for asset in new_run.assets:
            log.debug("link asset: %s -> %s", asset.original_path, asset.path)
            Path(asset.original_path).link(asset.path)
    except Exception as error:
        log.warn("linking error: %s -> %s", asset.original_path, asset.path)
        log.debug('cleaning up database')
        api.delete(root_path, new_run)
        manager.commit()
        raise error



# -*- coding: utf-8 -*-
import ruamel.yaml


def build_meta(case_name, segments):
    """Build metadata information content.

    Args:
        run (AnalysisRun): analysis run record
        qc_ped (path): path to qc pedigree file

    Returns:
        str: YAML-formatted string of the data
    """
    sample_map = {sample['sample_id']: sample['sample_name']
                  for sample in segments['pedigree']['samples']}
    metadata = {
        'name': case_name,
        'pipeline': 'mip',
        'pipeline_version': segments['family']['mip_version'],
        'analyzed_at': segments['family']['analysis_date'],
        'samples': sample_map,
    }
    return ruamel.yaml.dump(metadata, Dumper=ruamel.yaml.RoundTripDumper)



# -*- coding: utf-8 -*-
"""Archive a run to PDC.

1. get the archives for the run
2. send to PDC function (magic happens)
3. update new table with information about archive
4. remove the archive files
5. update that the case is archived
"""
from datetime import datetime
import logging

from .utils import on_pdc, send_to_pdc
from housekeeper.store import api

log = logging.getLogger(__name__)


def archive_run(run_obj, force):
    """Archive a run to PDC."""
    assert run_obj.compiled_at, "run not compiled yet"
    # fetch the archive assets
    assets = get_archiveassets(run_obj.id)
    # do PDC magic....
    # remove the archive assets
    for archive_type, asset in assets:
        if not on_pdc(asset.path) or force:
            send_to_pdc(asset.path)
        #log.debug("deleting: %s", asset.path)
        #api.delete_asset(asset)
    run_obj.compiled_at = None
    # mark the case as archived
    run_obj.archived_at = datetime.now()


def get_archiveassets(run_id):
    """Get the archive assets for a run."""
    for archive_type in ("data", "result"):
        category = "archive-{}".format(archive_type)
        query = api.assets(run_id=run_id, category=category)
        assets = query.all()
        for asset in assets:
            yield archive_type, asset



# -*- coding: utf-8 -*-
"""Restore archived files."""
import tarfile

from path import Path
import ruamel.yaml

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from .tar import untar_files
from .core import checksum


def restore_run(root_path, run_obj, tar_path, archive_type):
    """Restore files from a tar archive."""
    # confirm the checksum
    sha1 = get_checksum_archive(run_obj, tar_path, archive_type)
    #assert checksum(tar_path) == sha1

    # unpack the files
    run_dir = get_rundir(root_path, run_obj.case.name, run_obj)
    untar_files(run_dir, tar_path)

    # update the "local" status of all files blindly
    for asset in run_obj.assets:
        if Path(asset.path).exists():
            asset.is_local = True


def get_checksum_archive(run_obj, tar_path, archive_type):
    """Get the checksum for the correct Asset"""

    for asset in run_obj.assets:
        if asset.archive_type == archive_type and asset.basename() == Path(tar_path).basename():
            return asset.checksum


def run_fromtar(tar_path):
    """Get the run record for a tar archive."""
    with tarfile.open(tar_path, 'r:*') as in_handle:
        filename = 'meta.yaml'
        meta_file = in_handle.extractfile(filename)
        meta_data = ruamel.yaml.safe_load(meta_file)

    run_date = meta_data['analyzed_at'].date()
    query = api.runs(meta_data['name'], run_date=run_date)
    run_obj = query.first()
    return run_obj



# -*- coding: utf-8 -*-
from __future__ import division
import csv
import logging
import re

from path import Path
import ruamel.yaml

from housekeeper.exc import AnalysisNotFinishedError, UnsupportedVersionError
from .meta import build_meta, write_meta

MULTIQC_SAMTOOLS = 'multiqc/multiqc_data/multiqc_samtools.txt'

log = logging.getLogger(__name__)


def prepare_run(segments, force=False):
    """Prepare MIP output."""
    if not force:
        validate(segments['family'])

    outdata_dir = segments['config']['outDataDir']
    qcped_path = segments['family']['PedigreeFileAnalysis']['Path']
    fam_key = segments['config']['familyID']
    customer = segments['family']['InstanceTag'][0]
    case_name = "{}-{}".format(customer, fam_key)
    meta_output = build_meta(case_name, segments['family'], qcped_path)
    write_meta(meta_output, outdata_dir)

    qcmetrics_path = (segments['family']['Program']['QCCollect']['QCCollectMetricsFile']['Path'])
    sample_ids = segments['config']['sampleIDs']
    modify_qcmetrics(outdata_dir, qcmetrics_path, sample_ids)


def validate(family):
    """Validate analysis."""
    run_status = family['AnalysisRunStatus']
    if run_status != 'Finished':
        raise AnalysisNotFinishedError(run_status)

    version = family['MIPVersion']
    if not version.startswith('v3'):
        raise UnsupportedVersionError(version)


def modify_qcmetrics(outdata_dir, qcmetrics_path, sample_ids, multiqc_samtools=MULTIQC_SAMTOOLS):
    # multiqc outputs
    qc_samples = {}
    for sample_id in sample_ids:
        multiqc_path = Path(outdata_dir).joinpath(sample_id, multiqc_samtools)
        if multiqc_path.exists():
            log.debug("calculate total mapped reads for: %s", sample_id)
            with open(multiqc_path, 'r') as stream:
                mapped_data = total_mapped(stream)
                qc_samples[sample_id] = mapped_data
        else:
            log.warn("multiqc output missing for %s", sample_id)
            qc_samples[sample_id] = {}

        # duplicates
        log.debug("calculate duplicates for: %s", sample_id)
        duplicates = get_duplicates(outdata_dir, sample_id)
        qc_samples[sample_id]['duplicates'] = duplicates

    log.debug("parse QC metrics YAML: %s", qcmetrics_path)
    with open(qcmetrics_path, 'r') as stream:
        qc_data = ruamel.yaml.safe_load(stream)
        qc_rootkey = 'sample' if 'sample' in qc_data else list(qc_data.keys())[0]

    for sample_id, mapped_data in qc_samples.items():
        qc_data[qc_rootkey][sample_id]['MappedReads'] = mapped_data.get('mapped')
        qc_data[qc_rootkey][sample_id]['TotalReads'] = mapped_data.get('total')
        qc_data[qc_rootkey][sample_id]['MappedRate'] = mapped_data.get('percentage')
        qc_data[qc_rootkey][sample_id]['Duplicates'] = mapped_data['duplicates']

    new_qcmetrics = Path(qcmetrics_path.replace('.yaml', '.mod.yaml'))
    log.info("create updated qc metrics: %s", new_qcmetrics)
    with new_qcmetrics.open('w') as stream:
        ruamel.yaml.dump(qc_data, stream=stream, Dumper=ruamel.yaml.RoundTripDumper)


def total_mapped(stream):
    """Extract the percentage of mapped reads across all lanes."""
    data = csv.DictReader(stream, delimiter='\t', quoting=csv.QUOTE_NONE)
    total = 0
    mapped = 0
    for row in data:
        total += float(row.get('raw total sequences') or
                       row['raw_total_sequences'])
        mapped += float(row.get('reads mapped') or row['reads_mapped'])
    return {'mapped': mapped, 'total': total, 'percentage': mapped / total}


def get_duplicates(outdata_dir, sample_id):
    """Parse out the duplicate rate."""
    files = Path(outdata_dir).glob("{0}/bwa/{0}_lanes_*_sorted_md_metric"
                                   .format(sample_id))
    if len(files) == 1:
        with open(files[0], 'r') as stream:
            content = stream.read()
        dups = float(re.search('Fraction Duplicates: (.*)', content).groups()[0])
    else:
        raise IOError('no picard tools file!')
    return dups



# -*- coding: utf-8 -*-
from datetime import date as make_date
from subprocess import CalledProcessError
import logging

from housekeeper.store import api
from housekeeper.compile.utils import launch

log = logging.getLogger(__name__)


def on_pdc(file_path, date=None):
    """Check if a file is on PDC """

    try:
        log.debug('dsmc q archive {}'.format(file_path))
        launch('dsmc q archive {}'.format(file_path), envs={'DSM_DIR':'/opt/adsm_clinical'})
    except CalledProcessError as dsmc_exc:
        if dsmc_exc.returncode == 8:
            return False
        raise dsmc_exc
    return True


def send_to_pdc(file_path):
    """ Send a file to PDC"""
    log.info('sending: {}'.format(file_path))
    log.debug('dsmc archive {}'.format(file_path))
    launch('dsmc archive {}'.format(file_path), envs={'DSM_DIR':'/opt/adsm_clinical'})



# -*- coding: utf-8 -*-
import logging

import click
from path import Path

from housekeeper.store import api

log = logging.getLogger(__name__)


def migrate_root(manager, old_root, new_root, only_db=False):
    """Migrate root of analysis assets."""
    new_root = Path(new_root).abspath().normpath()
    log.debug("ensure parent of new root exists: %s", new_root.parent)
    new_root.parent.makedirs_p()
    if not only_db and new_root.exists():
        raise ValueError("new root can't exist")
    if not only_db:
        move_root(old_root, new_root)
    replace_paths(old_root, new_root)
    manager.commit()


def move_root(old_root, new_root):
    """Move the analysis assets from one directory to another."""
    Path(old_root).move(new_root)


def replace_paths(old_root, new_root):
    """Replace root paths for assets."""
    all_assets = api.assets()
    with click.progressbar(all_assets,
                           label='updating asset paths',
                           length=all_assets.count()) as bar:
        for asset in bar:
            new_path = asset.path.replace(old_root, new_root)
            log.debug("replacing path: %s -> %s", asset.path, new_path)
            assert Path(new_path).exists(), "can't find asset: {}".format(new_path)
            asset.path = new_path



# -*- coding: utf-8 -*-
import datetime
import os
import logging

import click
from dateutil.parser import parse as parse_date
from path import Path
import ruamel.yaml

from housekeeper.cli.utils import run_orabort
from housekeeper.constants import EXTRA_STATUSES
from housekeeper.store.models import ExtraRunData
from .migrate import migrate_root
from .utils import get_rundir
from . import api, Asset, AnalysisRun, Sample, Case

RUN_STATUSES = ['analyzed', 'compiled', 'delivered', 'answeredout', 'archived',
                'cleanedup']
SAMPLE_STATUSES = ['received', 'sequenced', 'confirmed']
CASE_STATUSES = RUN_STATUSES + EXTRA_STATUSES

log = logging.getLogger(__name__)


@click.command()
@click.option('-c', '--case', help='filter by case')
@click.option('-s', '--sample', help='filter by sample')
@click.option('-i', '--infer-case', is_flag=True,
              help='filter by related case for a sample')
@click.option('-t', '--category', help='filter by asset category')
@click.option('-a', '--all', 'all_runs', is_flag=True, default=False,
              help='get assets from ALL case runs, not only latest')
@click.pass_context
def get(context, case, sample, infer_case, category, all_runs):
    """Ask Housekeeper for a file."""
    api.manager(context.obj['database'])
    if sample and (case is None):
        sample_obj = api.sample(sample)
        if sample_obj is None:
            log.warn('sorry, sample not found')
            context.abort()
        case = sample_obj.case.name
        if infer_case:
            sample = None

    if not all_runs and (case or sample):
        # get assets only from latest run
        latest_run = api.runs(case_name=case).first()
        if latest_run is None:
            log.error("no analysis run found: %s", case)
            context.abort()
        run_id = latest_run.id
    else:
        # get assets from all runs
        run_id = None

    asset_query = api.assets(case_name=case, sample=sample, category=category,
                             run_id=run_id)
    if asset_query.count() == 0:
        log.error("no matching assets found")
        context.abort()
    else:
        for asset in asset_query:
            click.echo(asset.path)


@click.command()
@click.option('-b', '--before', help='list runs before a date')
@click.option('-a', '--after', help='list runs after a date')
@click.option('--archived/--no-archived', is_flag=True,
              help='list archived/not archived runs')
@click.option('--compiled/--no-compiled', is_flag=True,
              help='list compiled/not compiled runs')
@click.option('--cleaned/--no-cleaned', is_flag=True,
              help='list cleaned up/not cleaned up runs')
@click.option('--to-clean', is_flag=True,
              help='list runs ready to be cleaned up')
@click.option('-o', '--output', type=click.Choice(['root', '', 'sample', 'case']),
              default='case', help='what to output to console')
@click.option('-l', '--limit', default=20)
@click.argument('case_id', required=False)
@click.pass_context
def runs(context, before, after, archived, compiled, cleaned, to_clean,
         limit, output, case_id):
    """List runs loaded in the database."""
    api.manager(context.obj['database'])
    root_path = context.obj['root']
    before_date = parse_date(before) if before else None
    after_date = parse_date(after) if after else None
    run_q = api.runs(case_name=case_id, before=before_date, after=after_date,
                     archived=archived, compiled=compiled, cleaned=cleaned,
                     to_clean=to_clean).limit(limit)
    if run_q.first() is None:
        log.error("no runs found")
        context.abort()
    for run_obj in run_q:
        if output == 'root':
            run_root = get_rundir(root_path, run_obj.case.name, run_obj)
            click.echo(run_root)
        elif output == 'sample':
            for sample in run_obj.samples:
                click.echo(sample.name)
        else:
            click.echo("{} - {}".format(run_obj.case.name,
                                        run_obj.analyzed_at.date()))


@click.command()
@click.argument('asset_path')
@click.option('-s', '--short', is_flag=True, default=False,
              help='Print only the filename in the sha1sum file')
@click.option('-d', '--dash', is_flag=True, default=False,
              help='Print a dash instead of the filename')
@click.pass_context
def getsha1(context, asset_path, short, dash):
    """Ask Housekeeper for the sha1sum of a file."""
    api.manager(context.obj['database'])
    checksum = api.sha1(asset_path)
    checksum_of = asset_path
    if short:
        checksum_of = os.path.basename(asset_path)
    if dash:
        checksum_of = '-'

    # for some reason we need two spaces
    click.echo("  ".join((checksum, checksum_of)))


@click.command()
@click.option('-d', '--days', default=30, help='days to postpone clean up')
@click.argument('case_name')
@click.pass_context
def postpone(context, days, case_name):
    """Ask Housekeeper for a file."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(case_name)
    api.postpone(run_obj, time=datetime.timedelta(days=days))
    manager.commit()
    click.echo("analysis will be archived on: {}"
               .format(run_obj.will_cleanup_at))


@click.group()
@click.pass_context
def delete(context):
    """Delete things from the database."""
    pass


@delete.command('analysis')
@click.option('-y', '--yes', is_flag=True, help="skip manual confirmation")
@click.option('-d', '--date', help="date of the particular run")
@click.argument('case_name')
@click.pass_context
def delete_analysis(context, date, yes, case_name):
    """Delete an analysis run and files."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    root_path = context.obj['root']
    run_root = get_rundir(root_path, case_name, run_obj)
    click.echo("you are about to delete: {}".format(run_root))
    if yes or click.confirm('Are you sure?'):
        api.delete(root_path, run_obj)
        manager.commit()


@delete.command('sample')
@click.option('-y', '--yes', is_flag=True, help='skip confirmation')
@click.argument('sample_id')
@click.pass_context
def delete_sample(context, yes, sample_id):
    """Delete a sample from the database."""
    manager = api.manager(context.obj['database'])
    sample_obj = api.sample(sample_id)
    if sample_obj is None:
        log.error("sample not found: {}".format(sample_id))
        context.abort()
    click.echo("you are about to delete sample: {} - {}"
               .format(sample_obj.lims_id, sample_obj.id))
    if yes or click.confirm('Are you sure?'):
        sample_obj.delete()
        manager.commit()


@delete.command('case')
@click.option('-y', '--yes', is_flag=True, help='skip confirmation')
@click.argument('case_name')
@click.pass_context
def delete_case(context, yes, case_name):
    """Delete a case from the database."""
    manager = api.manager(context.obj['database'])
    case_obj = api.case(case_name)
    if case_obj is None:
        log.error("case not found: {}".format(case_name))
        context.abort()
    for sample_obj in case_obj.samples:
        log.warn("will remove sample: %s", sample_obj.lims_id)
    for analysis_run in case_obj.runs:
        log.warn("will remove analysis run: %s", analysis_run.analyzed_at.date())
    click.echo("you are about to delete case: {} - {}"
               .format(case_obj.name, case_obj.id))
    if yes or click.confirm('Are you sure?'):
        case_obj.delete()
        manager.commit()


@click.command()
@click.option('-f', '--force', is_flag=True,
              help='auto confirm and override sanity checks')
@click.option('-d', '--date', help="date of the particular run")
@click.option('-l', '--limit', default=20)
@click.option('-b', '--before', help='runs before a date')
@click.option('-a', '--after', help='runs after a date')
@click.option('-u', '--untagged-only', is_flag=True, default=False, help='only remove assets without archive type')
@click.argument('case_name', required=False)
@click.pass_context
def clean(context, force, date, limit, before, after, untagged_only, case_name):
    """Clean up files for an analysis.
    Providing a CASE_NAME will ignore --before, --after options
    """
    manager = api.manager(context.obj['database'])
    if case_name:
        run_objs = [run_orabort(context, case_name, date)]
    else:
        before_date = parse_date(before) if before else None
        after_date = parse_date(after) if after else None
        run_objs = api.runs(before=before_date, after=after_date).limit(limit)
        if run_objs.first() is None:
            log.error("no runs found")
            context.abort()

    if force or click.confirm('Cleaning following runs: {}'.
                              format([ run_obj.case.name for run_obj in run_objs ])):
        if untagged_only:
            force = True
        for run_obj in run_objs:
            click.echo('Cleaning run: {}'.format(run_obj))
            api.clean_up(context.obj['root'], run_obj, force=force,
                         untagged_only=untagged_only)
        manager.commit()


@click.command()
@click.option('-l', '--limit', default=20, help='limit number of results')
@click.option('-o', '--offset', default=0, help='skip initial results')
@click.option('-s', '--since', help='consider runs since date')
@click.option('--older', is_flag=True, help='reverse order of results')
@click.option('-c', '--category', default='bcf-raw',
              help='category of asset to return')
@click.pass_context
def ls(context, limit, offset, since, older, category):
    """List files from recently added runs."""
    api.manager(context.obj['database'])
    if category == 'samples':
        query = api.samples().join(Sample.run)
    elif category == 'cases':
        query = AnalysisRun.query
    else:
        query = api.assets(category=category).join(Asset.run)

    if older:
        query = query.order_by(AnalysisRun.analyzed_at)
    else:
        query = query.order_by(AnalysisRun.analyzed_at.desc())

    if since:
        date_obj = build_date(since) if since else None
        if older:
            query = query.filter(AnalysisRun.analyzed_at <= date_obj)
        else:
            query = query.filter(AnalysisRun.analyzed_at >= date_obj)
    else:
        query = query.offset(offset).limit(limit)

    if query.first() is None:
        log.warn('sorry, no matching assets found')
        context.abort()

    if category == 'samples':
        # we only consider unqiue samples
        output = (sample.lims_id for sample in query)
    elif category == 'cases':
        output = set(run.case.name for run in query)
    else:
        # we only consider files from the latest run per case
        cases = set()
        output = []
        for asset in query:
            if asset.run.case_id not in cases:
                output.append(asset.path)
                cases.add(asset.run.case_id)
    click.echo(" ".join(output))


@click.command()
@click.option('-y', '--yes', is_flag=True, help="answer 'yes' to confirmations")
@click.option('-o', '--only-db', is_flag=True, help='skip migrating assets')
@click.argument('new_root', type=click.Path())
@click.pass_context
def migrate(context, yes, only_db, new_root):
    """Migrate all assets from one root dir to another."""
    manager = api.manager(context.obj['database'])
    migrate_root(manager, context.obj['root'], new_root, only_db=only_db)

    if 'config' in context.obj:
        config_path = Path(context.obj['config'])
        if yes or click.confirm("update config file?"):
            # update the root in the config
            with config_path.open('w') as out_handle:
                ruamel.yaml.dump(context.obj, stream=out_handle,
                                 Dumper=ruamel.yaml.RoundTripDumper)


@click.command()
@click.option('-s', '--sample-status', type=click.Choice(SAMPLE_STATUSES),
              help='update sample status date')
@click.option('-r', '--run-status', type=click.Choice(RUN_STATUSES),
              help='update run status date')
@click.option('-c', '--extra-status', type=click.Choice(EXTRA_STATUSES),
              help='custom run status')
@click.option('-d', '--date', help='custom date')
@click.option('-n', '--now', is_flag=True, help='set date to "now"')
@click.option('-rd', '--run-date', help='date of a particular run')
@click.argument('identifier')
@click.pass_context
def status(context, date, now, run_date, sample_status, run_status,
           extra_status, identifier):
    """Mark dates for resources."""
    manager = api.manager(context.obj['database'])

    # determine which date to set: custom or now!
    if date:
        status_date = parse_date(date)
    elif now:
        status_date = datetime.datetime.now()
    else:
        status_date = None

    if sample_status:
        # update a date for a sample
        model_obj = api.sample(identifier)
        if model_obj is None:
            click.echo("can't find sample")
            context.abort()
        # types are predefined; no need to check!
        status_type = sample_status
    else:
        # update a date for an analysis run
        model_obj = run_orabort(context, identifier, run_date)
        # types are predefined as well
        status_type = run_status

    if extra_status:
        if model_obj.extra is None:
            log.info("adding extra data record")
            model_obj.extra = ExtraRunData()
            manager.commit()
        extra_date_key = "{}_date".format(extra_status)
        if status_date is None:
            # show a custom date for an analysis run
            current_date = getattr(model_obj.extra, extra_date_key)
        else:
            # update a custom date for a run
            setattr(model_obj.extra, extra_date_key, status_date)
            status_field = extra_status
    else:
        status_field = "{}_at".format(status_type)
        if status_date is None:
            # show a predfined date for a model
            current_date = getattr(model_obj, status_field)
        else:
            # update a predefined date on a model
            setattr(model_obj, status_field, status_date)

    if status_date is None:
        if current_date is None:
            context.abort()
        else:
            click.echo(current_date)
    else:
        log.info("updating %s -> %s", status_field, status_date)
        manager.commit()


@click.command()
@click.option('-l', '--limit', default=20, help='limit number of results')
@click.option('-o', '--offset', default=0, help='skip initial results')
@click.option('-m', '--missing', type=click.Choice(CASE_STATUSES))
@click.option('-v', '--version', help='filter on pipeline version')
@click.option('-r', '--ready', is_flag=True, help='DEPRECATED')
@click.pass_context
def cases(context, limit, offset, missing, version, ready):
    """Display information about cases."""
    api.manager(context.obj['database'])
    query = api.cases(missing=missing, version=version)
    for case in query.offset(offset).limit(limit):
        click.echo(case.name)


@click.command()
@click.option('-l', '--limit', default=20, help='limit number of results')
@click.option('-o', '--offset', default=0, help='skip initial results')
@click.option('-c', '--case', help='return samples related to a case')
@click.option('-m', '--missing', type=click.Choice(SAMPLE_STATUSES))
@click.pass_context
def samples(context, limit, offset, case, missing):
    """Display information about samples."""
    api.manager(context.obj['database'])
    status_to = {'received': 'receive', 'sequenced': 'sequence', 'confirmed': 'confirm'}
    query = api.samples(status_to=status_to.get(missing))

    if case:
        query = query.join(Sample.case).filter(Case.name == case)

    for sample in query.offset(offset).limit(limit):
        click.echo(sample.lims_id)


@click.command()
@click.option('-p', '--prioritize', is_flag=True, help='mark high priority')
@click.option('-f', '--flowcell', 'flowcells', multiple=True,
              help='link flowcell to sample')
@click.argument('sample_name')
@click.pass_context
def sample(context, prioritize, flowcells, sample_name):
    """Show or update a sample."""
    manager = api.manager(context.obj['database'])
    sample_obj = api.sample(sample_name)
    if sample_obj is None:
        log.error("no sample found with that name: %s", sample_name)
        context.abort()
    if prioritize:
        sample_obj.priority = True
        manager.commit()
        log.info("prioritized sample!")
    if flowcells:
        for flowcell in flowcells:
            if len(flowcell) != 9:
                log.error("invalid flowcell id: %s", flowcell)
                context.abort()
        flowcell_set = set(flowcells)
        flowcell_set.update(sample_obj.flowcells)
        sample_obj.flowcells = list(flowcell_set)
        manager.commit()
        log.info("updated sample flowcells: %s", sample_obj._flowcells)

    click.echo(sample_obj.to_json(pretty=True))


@click.command('add-sample')
@click.option('-d', '--date', help='date received for sample')
@click.option('-p', '--priority', is_flag=True, help='mark high priority')
@click.argument('customer')
@click.argument('family_id')
@click.argument('lims_id')
@click.pass_context
def add_sample(context, date, priority, customer, family_id, lims_id):
    """Add a new sample to the database."""
    manager = api.manager(context.obj['database'])
    existing_sample = Sample.query.filter_by(lims_id=lims_id).first()
    if existing_sample:
        log.error("sample already exists: %s", lims_id)
        context.abort()
    new_sample = Sample(lims_id=lims_id, priority=priority)
    if date:
        new_sample.received_at = parse_date(date)

    # also add a new case if it doesn't already exist
    case_name = '-'.join([customer, family_id])
    case_obj = api.case(case_name)
    if case_obj is None:
        case_obj = Case(name=case_name, customer=customer, family_id=family_id)
        manager.add(case_obj)
        log.info("added new case: %s", case_name)
    else:
        log.debug("case already exists: %s", case_name)

    # link sample to case
    case_obj.samples.append(new_sample)
    manager.commit()
    log.info("added new sample: %s", lims_id)


@click.command('add-case')
@click.argument('customer')
@click.argument('family_id')
@click.pass_context
def add_case(context, customer, family_id):
    """Add a case to the database."""
    manager = api.manager(context.obj['database'])
    case_name = "-".join([customer, family_id])
    existing_case = api.case(case_name)
    if existing_case:
        log.error("case already exists: %s", case_name)
        context.abort()
    new_case = Case(name=case_name, customer=customer, family_id=family_id)
    manager.add_commit(new_case)
    log.info("added new case: %s", case_name)


def build_date(date_str):
    """Parse date out of string."""
    return datetime.date(*map(int, date_str.split('-')))



# -*- coding: utf-8 -*-
from __future__ import division
import logging
import os

from housekeeper.exc import AnalysisNotFinishedError
from housekeeper.pipelines.mip.meta import write_meta, build_meta

log = logging.getLogger(__name__)


def prepare_run(segments, force=False):
    """Prepare MIP output."""
    if not force:
        validate(segments['family'])

    fam_key = segments['config']['familyID']
    mip_version = segments['family'].get('MIPVersion', 'v2.x')
    outdata_dir = segments['family']['ArchivePath']
    qcped_path = os.path.join(outdata_dir, fam_key, 'qc_pedigree.yaml')
    customer = segments['family']['InstanceTag'][0]
    case_name = "{}-{}".format(customer, fam_key)
    meta_output = build_meta(case_name, segments['family'], qcped_path,
                             version=mip_version, strict=False)
    write_meta(meta_output, outdata_dir)


def validate(family):
    """Validate analysis."""
    run_status = family['AnalysisRunStatus']
    if run_status != 'Archived':
        raise AnalysisNotFinishedError(run_status)



# -*- coding: utf-8 -*-
import logging

from path import Path
import ruamel.yaml

from housekeeper.exc import MissingFileError

log = logging.getLogger(__name__)


def parse_references(references, segments):
    """Combine the functions."""
    for reference in references:
        try:
            keys = reference['key'].split('|')
            source = segments[reference['source']]
            if reference.get('sample'):
                for sample_id, values in source.items():
                    ref_paths = parse_tree(values, keys)
                    for ref_path in ref_paths:
                        ref_path = format_path(reference, ref_path)
                        yield {'reference': reference, 'path': ref_path,
                               'sample': sample_id}
            else:
                ref_paths = parse_tree(source, keys)
                for ref_path in ref_paths:
                    ref_path = format_path(reference, ref_path)
                    yield {'reference': reference, 'path': ref_path}
        except (KeyError, MissingFileError) as error:
            if reference.get('required') is False:
                log.warn(error.message)
                continue
            else:
                log.error("source: {}, key: {}".format(reference['source'],
                                                       reference['key']))
                raise error


def prepare_inputs(config_data):
    """Parse input and pick out segments."""
    fam_key = config_data['familyID']
    sampleinfo_path = config_data['sampleInfoFile']
    log.debug("parse sampleinfo YAML: %s", sampleinfo_path)
    with open(sampleinfo_path, 'r') as in_handle:
        data = ruamel.yaml.safe_load(in_handle)
    family = data[fam_key][fam_key]
    samples = {key: subdata for key, subdata in data[fam_key].items()
               if key != fam_key}
    return {'family': family, 'samples': samples, 'config': config_data}


def format_path(reference, orig_path):
    """Get a file from analysis output."""
    if 'suffix' in reference:
        new_path = orig_path + reference['suffix']
    elif 'replace' in reference:
        new_path = orig_path.replace(reference['replace']['old_str'],
                                     reference['replace']['new_str'])
    elif 'glob' in reference:
        matches = Path(orig_path).glob(reference['glob'])
        if len(matches) == 0:
            glob = reference['glob']
            raise MissingFileError("missing: {}, {}".format(orig_path, glob))
        new_path = matches[0]
    else:
        new_path = orig_path
    return new_path


def parse_tree(tree, keys):
    """Parse a dict using a list of keys."""
    for index, key in enumerate(keys):
        if key.isdigit():
            if isinstance(tree, dict):
                tree = list(tree.values())[int(key)]
            else:
                tree = tree[int(key)]
        elif key == '*':
            subtrees = tree.values()
            paths = []
            for subtree in subtrees:
                values = parse_tree(subtree, keys[index + 1:])
                for value in values:
                    paths.append(value)
            return paths
        else:
            tree = tree[key]
    return [tree]



# -*- coding: utf-8 -*-
import pkg_resources
from collections import namedtuple
from datetime import datetime
import hashlib
import logging
import itertools

from path import Path

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from .utils import launch
from .tar import tar_files

BLOCKSIZE = 65536
ArchiveGroup = namedtuple('ArchiveGroup', ['id', 'out', 'checksum'])
log = logging.getLogger(__name__)


def compile_run(root_path, run_obj):
    """High level compile function for a run."""
    groups = compress_run(root_path, run_obj)
    for group in groups:
        log.info("compress %s archive: %s", group.id, group.out)
        category = "archive-{}".format(group.id)
        new_asset = api.add_asset(run_obj, group.out, category)
        new_asset.path = group.out
        new_asset.checksum = group.checksum
        new_asset.archive_type = 'archive'
        run_obj.assets.append(new_asset)
    run_obj.compiled_at = datetime.now()


def compress_run(root_path, run_obj):
    """Archive a run into data+results archives."""
    run_dir = Path(get_rundir(root_path, run_obj.case.name, run=run_obj))
    for group, paths in group_assets(run_obj.assets):
        run_date = run_obj.analyzed_at.date()
        group_out = run_dir.joinpath("{}.{}.tar.gz".format(run_date, group))
        filenames = [Path(full_path).basename() for full_path in paths]
        tar_files(group_out, root_dir=run_dir, filenames=filenames)
        sha1 = checksum(group_out)
        yield ArchiveGroup(id=group, out=group_out, checksum=sha1)


def encrypt_run(root_path, run_obj):
    """ Encrypts an input file and places the encrypted archive
    in the run dir.
    """
    run_dir = Path(get_rundir(root_path, run_obj.case.name, run=run_obj))
    groups = ('data', 'result')
    archives = []
    for group in groups:
        category = "archive-{}".format(group)
        archive_grp = api.assets(run_id=run_obj.id, category=category).first()
        if archive_grp:
            archives.append(archive_grp)


    encrypt_script = pkg_resources.resource_filename('housekeeper', 'scripts/encrypt.batch')
    for asset in itertools.chain(archives):

        # here we go: encrypt!
        stdout = launch('{} {} {}'.format(encrypt_script, asset.path, run_dir))

        # the rest is administration
        archive_path = '{}.gpg'.format(asset.path)
        log.info(archive_path)

        compilation_path = asset.path

        new_asset = api.add_asset(run_obj, archive_path, asset.category)
        new_asset.path = archive_path
        new_asset.archive_type = 'archive'
        new_asset.checksum = checksum(archive_path)
        run_obj.assets.append(new_asset)

        api.delete_asset(asset)
    run_obj.compiled_at = datetime.now()


def group_assets(assets):
    """Groups asset paths based on archive type."""
    # group files base on type
    groups = ('data', 'result')
    for group in groups:
        paths = [asset.path for asset in assets
                 if asset.archive_type in (group, 'meta')]
        yield group, paths


def checksum(path):
    """Calculcate checksum for a file."""
    hasher = hashlib.sha1()
    with open(path, 'rb') as stream:
        buf = stream.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = stream.read(BLOCKSIZE)
    return hasher.hexdigest()



# -*- coding: utf-8 -*-
import logging

from housekeeper.store import AnalysisRun
from housekeeper.constants import TIME_TO_CLEANUP

log = logging.getLogger(__name__)


def analysis(pipeline, version, analyzed_at, reference_genome=None):
    """Store information about a general analysis run.

    This is the most low level implementation of how to store files from an
    analysis.
    """
    new_run = AnalysisRun(pipeline=pipeline, pipeline_version=version,
                          analyzed_at=analyzed_at, reference_genome=reference_genome)

    # set the future date for archiving
    new_run.will_cleanup_at = analyzed_at + TIME_TO_CLEANUP
    return new_run



# -*- coding: utf-8 -*-
from functools import partial

import click
import ruamel.yaml

from housekeeper.store import api
from housekeeper.parsemip import parse_mipsex


@click.command()
@click.argument('case_id')
@click.argument('existing_data', type=click.File('r'), default='-')
@click.pass_context
def export(context, case_id, existing_data):
    """Export data about a case/run."""
    api.manager(context.obj['database'])
    run_obj = api.runs(case_id).first()
    hk_data = export_data(run_obj)
    existing = ruamel.yaml.safe_load(existing_data)
    new_data = fillin_data(existing, hk_data)
    raw_dump = ruamel.yaml.safe_dump(new_data, Dumper=ruamel.yaml.RoundTripDumper)
    click.echo(raw_dump)


def export_data(run_obj):
    """Export data about a case."""
    get_assets = partial(api.assets, run_id=run_obj.id)
    qc_metrics = api.assets(category='qc').first()

    family_data = {
        # taboo
        'bcf_raw': get_assets(category='gbcf').first().path,

        # loqusdb
        'pedigree': get_assets(category='pedigree').first().path,
        'vcf_research': get_assets(category='vcf-research-bin').first().path,

        # cgstats
        'sampleinfo': get_assets(category='sampleinfo').first().path,
        'qc': qc_metrics.path,
    }

    # get predicted sex
    qcm_data = ruamel.yaml.safe_load(open(qc_metrics.path, 'r'))
    predicted_sexes = parse_mipsex(qcm_data)

    samples_data = {
        sample.lims_id: {
            # taboo
            'predicted_sex': predicted_sexes[sample.lims_id],

            # chanjo
            'coverage': get_assets(sample=sample.lims_id,
                                   category='coverage').first().path,
        } for sample in run_obj.samples
    }

    return {'family': family_data, 'samples': samples_data}


def fillin_data(existing_data, hk_data):
    """Fill in data about a case with Housekeeper data."""
    # fill in family data
    existing_data.update(hk_data['family'])

    # fill in sample data
    for sample_data in existing_data['samples']:
        new_data = hk_data['samples'][sample_data['id']]
        sample_data.update(new_data)

    return existing_data



# -*- coding: utf-8 -*-
from copy import deepcopy
import datetime
import pytest


@pytest.fixture(scope='function')
def bundle_data():
    data = {
        'name': 'sillyfish',
        'created': datetime.datetime.now(),
        'expires': datetime.datetime.now(),
        'files': [{
            'path': 'tests/fixtures/example.vcf',
            'archive': False,
            'tags': ['vcf', 'sample']
        }, {
            'path': 'tests/fixtures/family.vcf',
            'archive': True,
            'tags': ['vcf', 'family']
        }]
    }
    return data


@pytest.fixture(scope='function')
def bundle_data_json(bundle_data):
    json_data = deepcopy(bundle_data)
    json_data['created'] = json_data['created'].isoformat()
    json_data['expires'] = json_data['expires'].isoformat()
    return json_data



# -*- coding: utf-8 -*-
from datetime import date as make_date
import logging

from path import Path

log = logging.getLogger(__name__)


def get_rundir(root_path, case_name, run=None):
    """Return path to run root dir."""
    case_root = Path(root_path).joinpath(case_name)
    if run:
        run_root = case_root.joinpath(run.analysis_date.isoformat())
        return run_root
    else:
        return case_root


def build_date(date_str):
    """Parse date out of string."""
    return make_date(*map(int, date_str.split('-')))



# -*- coding: utf-8 -*-
import datetime
import logging

from alchy import Manager
from path import Path
import sqlalchemy as sqa

from housekeeper.constants import EXTRA_STATUSES
from .models import AnalysisRun, Asset, Case, Model, Sample, ExtraRunData
from .utils import get_rundir

log = logging.getLogger(__name__)
DATE_2017 = datetime.datetime(2017, 1, 1)


def manager(db_uri):
    log.debug('open connection to database: %s', db_uri)
    db = Manager(config=dict(SQLALCHEMY_DATABASE_URI=db_uri), Model=Model,
                 session_options=dict(autoflush=False))
    return db


def assets(case_name=None, sample=None, category=None, run_id=None):
    """Get files from the database."""
    query = Asset.query
    if run_id:
        query = query.filter_by(run_id=run_id)
    elif case_name:
        query = (query.join(Asset.run, AnalysisRun.case)
                      .filter(Case.name == case_name))
    if sample:
        query = query.join(Asset.sample).filter(Sample.lims_id == sample)
    if category:
        query = query.filter(Asset.category == category)
    return query


def case(name):
    """Get a case from the database.

    Args:
        name (str): the unique name of the analysis

    Returns:
        Case or None: the case or None if not found
    """
    # 'name' is a unique column so it will always return 1 or 0 records
    case_obj = Case.query.filter_by(name=name).first()
    return case_obj


def samples(query_str=None, status_to=None, customer=None, family_id=None):
    """Get samples from the database."""
    query = (Sample.query.join(Sample.case)
                         .filter(Case.is_manual == None)
                         .order_by(Sample.priority.desc(), Sample.received_at))
    if query_str:
        query = query.filter(Sample.lims_id.like("%{}%".format(query_str)))

    if status_to == 'sequence':
        query = query.filter(Sample.received_at > datetime.date(2017, 1, 1),
                             Sample.sequenced_at == None)
    elif status_to == 'confirm':
        query = query.filter(Sample.sequenced_at != None, Sample.confirmed_at == None)
    elif status_to == 'receive':
        query = query.filter(Sample.received_at == None, Sample.sequenced_at == None)

    if customer:
        query = query.filter_by(customer=customer)
    if family_id:
        query = query.filter_by(family_id=family_id)

    return query


def sample(lims_id):
    """Get sample from the database."""
    sample_obj = Sample.query.filter_by(lims_id=lims_id).first()
    return sample_obj


def cases(query_str=None, missing=None, version=None, onhold=None):
    """Get multiple cases from the database."""
    query = Case.query.filter(Case.is_manual == None).order_by(Case.created_at)

    if missing == 'analyzed':
        not_analyzed = ((Sample.sequenced_at > datetime.date(2017, 1, 1)) &
                        (AnalysisRun.analyzed_at == None))
        query = (query.outerjoin(Case.runs)
                      .join(Case.samples)
                      .filter(not_analyzed | Case.rerun_requested))
    elif missing == 'delivered':
        query = (query.join(Case.runs)
                      .filter(AnalysisRun.analyzed_at != None,
                              AnalysisRun.delivered_at == None,
                              AnalysisRun.pipeline_version.like("%v4.%")))
    elif missing == 'archived':
        query = (query.join(Case.runs)
                      .filter(AnalysisRun.archived_at == None))
    elif missing == 'answeredout':
        query = (query.join(Case.runs)
                      .filter(AnalysisRun.delivered_at != None,
                              AnalysisRun.answeredout_at == None))
    elif missing == 'cleanedup':
        today = datetime.datetime.today()
        query = (query.join(Case.runs)
                      .filter(AnalysisRun.will_cleanup_at < today))
    elif missing in EXTRA_STATUSES:
        date_field = getattr(ExtraRunData, "{}_date".format(missing))
        query = (query.join(Case.runs, AnalysisRun.extra)
                      .filter(AnalysisRun.analyzed_at != None,
                              date_field == None))
    if version:
        like_str = "{}%".format(version)
        query = (query.join(Case.runs)
                      .filter(AnalysisRun.pipeline_version.like(like_str)))
    if query_str:
        query = query.filter(Case.name.like("%{}%".format(query_str)))

    if onhold is not None:
        query = query.filter(Case.is_onhold == onhold)
    return query


def runs(case_name=None, run_date=None, since=None, before=None, after=None,
         archived=None, compiled=None, cleaned=None, to_clean=None):
    """Get runs for a case from the database.

    Args:
        case_name (str): the unique name of the case

    Returns:
        Query: all runs for the case
    """
    run_query = AnalysisRun.query
    if run_date:
        delta = datetime.timedelta(days=1)
        next_day = run_date + delta
        condition = AnalysisRun.analyzed_at.between(run_date, next_day)
        run_query = run_query.filter(condition)

    order = AnalysisRun.analyzed_at.desc()
    if after:
        since = after
    if since:
        order = AnalysisRun.analyzed_at.desc()
        run_query = run_query.filter(AnalysisRun.analyzed_at > since)
    if before:
        order = AnalysisRun.analyzed_at
        run_query = run_query.filter(AnalysisRun.analyzed_at < before)

    if to_clean:
        today = datetime.datetime.today()
        run_query = run_query.filter(AnalysisRun.archived_at != None,
                                     AnalysisRun.cleanedup_at == None,
                                     AnalysisRun.will_cleanup_at < today)
    else:
        if archived:
            if archived is True:
                run_query = run_query.filter(AnalysisRun.archived_at != None)
            else:
                run_query = run_query.filter(AnalysisRun.archived_at == None)
        if compiled is not None:
            if compiled is True:
                run_query = run_query.filter(AnalysisRun.compiled_at != None)
            else:
                run_query = run_query.filter(AnalysisRun.compiled_at == None)
        if cleaned:
            if cleaned is True:
                run_query = run_query.filter(AnalysisRun.cleanedup_at != None)
            else:
                run_query = run_query.filter(AnalysisRun.cleanedup_at == None)
    if case_name:
        run_query = (run_query.join(AnalysisRun.case)
                              .filter(Case.name == case_name))
    return run_query.order_by(order)


def delete(root_path, run_obj):
    """Delete an analysis run along with related files on the system.

    Args:
        run_obj (AnalysisRun): the analysis run to delete
    """
    run_dir = get_rundir(root_path, run_obj.case.name, run_obj)
    delete_dir(run_dir)
    run_obj.delete()


def delete_asset(asset_obj):
    """Delete an asset completely from the database."""
    if asset_obj.is_local:
        Path(asset_obj.path).remove_p()
    asset_obj.delete()


def clean_up(root_path, run_obj, force=False, untagged_only=False):
    """Clean up files for an analysis."""
    # check if run is ready
    if not force and not all([run_obj.archived_at, run_obj.delivered_at]):
        # not ready for clean up!
        log.warn("run not ready for clean up")
    else:
        for asset in run_obj.assets:
            if not asset.archive_type:
                asset.delete()
                delete_file(asset.path)
            else:
                if not untagged_only:
                    asset.is_local = False

        if not untagged_only:
            run_dir = get_rundir(root_path, run_obj.case.name, run_obj)
            log.info("removing rundir: %s", run_dir)
            delete_dir(run_dir)
            run_obj.cleanedup_at = datetime.datetime.now()


def postpone(run_obj, time=datetime.timedelta(days=30)):
    """Postpone the automatic archival of analysis by X time."""
    run_obj.will_cleanup_at += time


def delete_file(filename):
    file_path = Path(filename)
    if file_path.exists():
        log.info("removing file : %s", filename)
        file_path.remove_p()
    else:
        log.debug("asset not found: %s", filename)


def delete_dir(directory):
    directory_path = Path(directory)
    if directory_path.exists():
        log.info("removing files under: %s", directory)
        directory_path.rmtree_p()
    else:
        log.debug("run directory not found: %s", directory)


def add_asset(run_obj, asset_path, category, archive_type=None, sample=None):
    """Link an asset to a run."""
    abs_path = Path(asset_path).abspath()
    new_asset = Asset(original_path=abs_path, category=category,
                      archive_type=archive_type)
    new_asset.sample = sample
    return new_asset


def sha1(asset_path):
    """Retrieves the sha1sum from an asset"""
    abs_path = Path(asset_path).abspath()
    query = Asset.query
    checksum = query.filter(Asset.original_path == abs_path).first().checksum

    return checksum


def to_analyzed(session, category, limit=50):
    """Calculate times it takes to analyze a sample."""
    query = (session.query(Sample.lims_id, Sample.received_at, AnalysisRun.analyzed_at)
                    .join(Sample.runs)
                    .filter(Sample.category == category, Sample.received_at > DATE_2017)
                    .order_by(Sample.received_at.desc())
                    .limit(limit))
    diffs = [{
        'name': item[0],
        'y': (item[2] - item[1]).days
    } for item in query]
    return diffs


def to_sequenced(session, category, limit=50):
    """Calculate time it takes to sequence samples."""
    day_diff = sqa.func.TIMESTAMPDIFF(sqa.text('DAY'), Sample.received_at, Sample.sequenced_at)
    query = (session.query(Sample.lims_id.label('name'), day_diff.label('y'))
                    .filter(Sample.category == category, Sample.received_at > DATE_2017)
                    .order_by(Sample.received_at.desc())
                    .limit(limit))
    diffs = [{'name': sample.name, 'y': sample.y} for sample in query]
    return diffs



# -*- coding: utf-8 -*-
"""
Rsync module based on:
https://gist.github.com/aerickson/1283442
"""
import logging
import re
import subprocess

log = logging.getLogger(__name__)


def sync_files(local_dir, remote_dir):
    """Sync a directory of files to a remote server."""
    # first dry-run to find out how many files to sync
    cmd = "rsync -az --stats --dry-run {} {}".format(local_dir, remote_dir)
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE)
    remainder = proc.communicate()[0]
    groups = re.findall(r'Number of files: (\d+)', remainder)
    total_files = int(groups[0])
    log.info("number of files: %s", total_files)

    # now do the real rsync
    cmd = "rsync -avz --progress {} {}".format(local_dir, remote_dir)
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE)

    while True:
        output = proc.stdout.readline()
        if 'to-check' in output:
            groups = re.findall(r'to-check=(\d+)/(\d+)', output)
            progress = (100 * (int(groups[0][1]) - int(groups[0][0]))) / total_files
            log.info("done: {}%".format(progress))
            if int(groups[0][0]) == 0:
                break


def uppmax(user, project_id, local_dir):
    """Sync files to UPPMAX project using rsync."""
    remote_dir = "{}@milou.uppmax.uu.se:/proj/{}/INBOX".format(user, project_id)
    sync_files(local_dir, remote_dir)



# -*- coding: utf-8 -*-
import logging
import subprocess
import os

log = logging.getLogger(__name__)



def launch(command, envs=None):
    """ Launch a command

    Args:
        command (str): the bash command to launch

    Returns (str): stdout
    """

    if envs:
        os.environ.update(envs)

    try:
        command_line = command.split(' ')
        log.debug(' '.join(command_line))
        log.info(command_line)
        return subprocess.check_output(command_line)
    except subprocess.CalledProcessError as exception:
        exception.message = "Command '{}' failed.".format(' '.join(command_line))
        raise exception



# -*- coding: utf-8 -*-
from flask import current_app
from flask_dance.contrib.google import make_google_blueprint
from flask_dance.consumer import oauth_authorized
from flask import flash, redirect, request, session, url_for
from flask_login import (LoginManager, login_user, login_required, logout_user,
                         AnonymousUserMixin, UserMixin)
from sqlalchemy import types, Column
from path import Path
import ruamel.yaml


class AnonymousUser(AnonymousUserMixin):

    def __init__(self):
        self.name = 'Paul T. Anderson'
        self.email = 'pt@anderson.com'

    @property
    def is_authenticated(self):
        if current_app.config.get('LOGIN_DISABLED'):
            return True
        else:
            return False


class UserManagementMixin(UserMixin):
    id = Column(types.Integer, primary_key=True)
    google_id = Column(types.String(128), unique=True)
    email = Column(types.String(128), unique=True)
    name = Column(types.String(128))
    avatar = Column(types.Text)

    @property
    def first_name(self):
        """First part of name."""
        return self.name.split(' ')[0]


class UserManagement(object):

    """Provide usermanagement for a Flask app.

    ENV variables:
    GOOGLE_OAUTH_CLIENT_ID
    GOOGLE_OAUTH_CLIENT_SECRET
    USER_DATABASE_PATH
    """

    def __init__(self, manager, User):
        super(UserManagement, self).__init__()
        self.User = User
        self.manager = manager
        self.login_manager = None
        self.blueprint = None
        self.setup()

    def init_app(self, app):
        app.register_blueprint(self.blueprint, url_prefix='/login')

        @app.route('/login')
        def login():
            """Redirect to the Google login page."""
            # store potential next param URL in the session
            if 'next' in request.args:
                session['next_url'] = request.args.get('next')
            return redirect(url_for('google.login'))

        @app.route('/logout')
        @login_required
        def logout():
            logout_user()
            flash('You have logged out', 'info')
            return redirect(url_for('index'))

        self.login_manager.init_app(app)

    def setup(self):
        self.blueprint = make_google_blueprint(scope=['profile', 'email'])

        # setup login manager
        self.login_manager = LoginManager()
        self.login_manager.login_view = 'login'
        self.login_manager.anonymous_user = AnonymousUser

        @self.login_manager.user_loader
        def load_user(user_id):
            return self.User.query.get(int(user_id))

        @oauth_authorized.connect_via(self.blueprint)
        def google_loggedin(blueprint, token, this=self):
            """Create/login local user on successful OAuth login."""
            if not token:
                flash("Failed to log in with {}".format(blueprint.name), 'danger')
                return redirect(url_for('index'))

            # figure out who the user is
            resp = blueprint.session.get('/oauth2/v1/userinfo?alt=json')

            if resp.ok:
                userinfo = resp.json()

                # check if the user is whitelisted
                email = userinfo['email']
                user_obj = this.User.query.filter_by(email=email).first()
                if user_obj is None:
                    flash("email not whitelisted: {}".format(email), 'danger')
                    return redirect(url_for('index'))

                if user_obj:
                    user_obj.name = userinfo['name']
                    user_obj.avatar = userinfo['picture']
                    user_obj.google_id = userinfo['id']
                else:
                    user_obj = this.User(google_id=userinfo['id'], name=userinfo['name'],
                                         avatar=userinfo['picture'], email=email)
                    this.manager.add(user_obj)

                this.manager.commit()
                login_user(user_obj)
                flash('Successfully signed in with Google', 'success')
            else:
                message = "Failed to fetch user info from {}".format(blueprint.name)
                flash(message, 'danger')

            next_url = session.pop('next_url', None)
            return redirect(next_url or url_for('index'))


class UserAdmin(object):

    """docstring for UserAdmin"""

    def __init__(self, database_path=None):
        super(UserAdmin, self).__init__()
        self.database_path = Path(database_path) if database_path else None

    def init_app(self, app):
        """Initialize in Flask app context."""
        self.database_path = Path(app.config['USER_DATABASE_PATH'])

    def confirm(self, email):
        """Confirm that a user has been whitelisted."""
        # read in the file on every request
        with self.database_path.open('r') as in_handle:
            whitelisted_emails = ruamel.yaml.safe_load(in_handle)['whitelist']
        return email in whitelisted_emails



# -*- coding: utf-8 -*-


class HousekeeperError(Exception):

    def __init__(self, message=None):
        self.message = message


class AnalysisConflictError(HousekeeperError):
    pass


class MissingFileError(HousekeeperError):
    pass


class AnalysisNotFinishedError(HousekeeperError):
    pass


class UnsupportedVersionError(HousekeeperError):
    pass


class MalformattedPedigreeError(HousekeeperError):
    pass



# -*- coding: utf-8 -*-
import logging

import ruamel.yaml

log = logging.getLogger(__name__)


def prepare_inputs(config_data):
    """Parse input and pick out segments."""
    sampleinfo_path = config_data['qccollect_sampleinfo_file']
    log.debug("parse sampleinfo YAML: %s", sampleinfo_path)
    with open(config_data['pedigree_file'], 'r') as in_handle:
        ped_data = ruamel.yaml.safe_load(in_handle)
    with open(sampleinfo_path, 'r') as in_handle:
        sampleinfo_data = ruamel.yaml.safe_load(in_handle)
    family = sampleinfo_data
    samples = {sample_id: sample_data for sample_id, sample_data in
               sampleinfo_data['sample'].items()}
    return {'family': family, 'samples': samples, 'config': config_data,
            'pedigree': ped_data}



# -*- coding: utf-8 -*-
import logging

import click

from housekeeper.store import api, AnalysisRun
from housekeeper.compile.cli import compile as compile_cmd
from .core import archive_run
from housekeeper.cli.utils import run_orabort

log = logging.getLogger(__name__)


@click.command()
@click.option('-d', '--date', help="date of the particular run")
@click.option("-f", "--force", is_flag=True, help="skip checks")
@click.argument('case_name')
@click.pass_context
def archive(context, date, force, case_name):
    """Archive compiled assets to PDC."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    if not run_obj.compiled_at:
        log.warn("run not compiled yet: %s", case_name)
        if force or click.confirm("do you want to compile the run?"):
            date_str = run_obj.analyzed_at.date().isoformat()
            context.invoke(compile_cmd, date=date_str, force=True,
                           case_name=case_name)
            # refresh the run object
            run_obj = AnalysisRun.query.get(run_obj.id)
        else:
            context.abort()

    if force or click.confirm("are you sure you want to archive the run?"):
        archive_run(run_obj, force=force)
        manager.commit()



# -*- coding: utf-8 -*-
from contextlib import contextmanager
import os
import tarfile


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def tar_files(out_file, root_dir, filenames):
    """Group files into a tar archive."""
    with cd(root_dir):
        with tarfile.open(out_file, mode='w:gz') as out_handle:
            for filename in filenames:
                out_handle.add(filename)


def untar_files(out_dir, tar_file):
    """Ungroup tar-ed filed into a directory."""
    with tarfile.open(tar_file, mode='r:*') as in_handle:
        in_handle.extractall(out_dir)



# -*- coding: utf-8 -*-
from __future__ import division
import logging

from path import Path
import ruamel.yaml

from housekeeper.exc import AnalysisNotFinishedError, UnsupportedVersionError
from housekeeper.pipelines.mip.meta import write_meta
from .meta import build_meta

log = logging.getLogger(__name__)


def prepare_run(segments, force=False):
    """Prepare MIP output."""
    if not force:
        validate(segments['family'])

    outdata_dir = segments['config']['outdata_dir']
    fam_key = segments['config']['family_id']
    customer = segments['pedigree']['owner']
    case_name = "{}-{}".format(customer, fam_key)
    meta_output = build_meta(case_name, segments)
    write_meta(meta_output, outdata_dir)
    qcmetrics_path = (segments['family']['program']['qccollect']
                              ['qccollect_metrics_file']['path'])
    modify_qcmetrics(outdata_dir, qcmetrics_path)


def validate(family):
    """Validate analysis."""
    run_status = family['analysisrunstatus']
    if run_status != 'finished':
        raise AnalysisNotFinishedError(run_status)

    version = family['mip_version']
    if not version.startswith('v4'):
        raise UnsupportedVersionError(version)


def modify_qcmetrics(outdata_dir, qcmetrics_path):
    """Summarize some stats on sample level."""
    with open(qcmetrics_path, 'r') as in_handle:
        qc_data = ruamel.yaml.safe_load(in_handle)

    for sample_id, values in qc_data['sample'].items():
        # extract data dicts
        datas = [data for level_id, data in values.items() if
                 level_id[-6:-1] == '.lane']
        reads, mapped = 0, 0
        for data in datas:
            if 'bamstats' in data:
                reads += data['bamstats']['raw_total_sequences']
                mapped += data['bamstats']['reads_mapped']
            else:
                log.warn("bamstats missing for lane, skipping")

        values['reads'] = reads
        values['reads_mapped'] = mapped
        values['reads_mapped_rate'] = mapped / reads

    new_qcmetrics = Path(qcmetrics_path.replace('.yaml', '.mod.yaml'))
    log.info("create updated qc metrics: %s", new_qcmetrics)
    with new_qcmetrics.open('w') as out_handle:
        ruamel.yaml.dump(qc_data, stream=out_handle, Dumper=ruamel.yaml.RoundTripDumper)



# -*- coding: utf-8 -*-
import logging

import click

from housekeeper.archive import get_archiveassets
from housekeeper.cli.utils import run_orabort
from housekeeper.store import api
from .core import compile_run, encrypt_run
from .restore import restore_run, run_fromtar

log = logging.getLogger(__name__)


@click.command()
@click.option('-d', '--date', help="date of the particular run")
@click.option("-f", "--force", is_flag=True, help="skip checks")
@click.option("-e", "--encrypt", is_flag=True, help="encrypt compilation")
@click.argument('case_name')
@click.pass_context
def compile(context, date, force, encrypt, case_name):
    """Compile assets into archives."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    if run_obj.compiled_at:
        if force or click.confirm("replace existing archives?"):
            for _, asset in get_archiveassets(run_obj.id):
                if asset:
                    log.debug("deleting: %s", asset.path)
                    api.delete_asset(asset)
            manager.commit()

    if force or click.confirm('Are you sure?'):
        compile_run(context.obj['root'], run_obj)
        manager.commit()
        if encrypt:
            encrypt_run(context.obj['root'], run_obj)
        manager.commit()


@click.command()
@click.argument('archive_type', type=click.Choice(['data', 'result']))
@click.argument('tar_archive', type=click.Path(exists=True))
@click.pass_context
def restore(context, archive_type, tar_archive):
    """Restore analysis run files."""
    manager = api.manager(context.obj['database'])
    run_obj = run_fromtar(tar_archive)
    if run_obj is None:
        log.error('no matching analysis run found')
        context.abort()
    restore_run(context.obj['root'], run_obj, tar_archive, archive_type)
    manager.commit()



# -*- coding: utf-8 -*-
import logging

import click
from path import Path

from housekeeper.store import api

log = logging.getLogger(__name__)


def setup_db(db_uri, reset=False):
    """Setup database with tables and store config.

    Args:
        root_path (path): root to store analyses
        db_uri (str): connection string to database
        reset (Optional[bool]): whether to reset an existing database
    """
    log.info("setup a new database: %s", db_uri)
    db = api.manager(db_uri)
    if reset:
        db.drop_all()
    db.create_all()


@click.command()
@click.option('--db-only', is_flag=True)
@click.option('--reset', is_flag=True)
@click.argument('root', type=click.Path())
@click.pass_context
def init(context, db_only, reset, root):
    """Setup the housekeeper."""
    root_path = Path(root).abspath()
    if not db_only:
        if root_path.exists():
            log.error("root path already exists: %s", root_path)
            context.abort()
        else:
            log.info("create root folder to store analyses: %s", root)
            root_path.makedirs_p()
    uri = context.obj.get('database')
    db_uri = uri or "sqlite:///{}".format(root_path.joinpath('store.sqlite3'))
    setup_db(db_uri, reset=reset)



# -*- coding: utf-8 -*-
from __future__ import division
import logging

from path import Path

from housekeeper.pipelines.general.add import analysis as general_analysis
from housekeeper.store import Asset

log = logging.getLogger(__name__)


def build_analysis(segments, version=None, customer=None):
    """Prepare info for a MIP analysis."""
    fam_key = segments['config']['familyID']
    version = version if version else segments['family']['MIPVersion']
    analyzed_at = segments['family']['AnalysisDate']
    sample_ids = segments['config']['sampleIDs']
    customer = customer or segments['family']['InstanceTag'][0]
    name = "{}-{}".format(customer, fam_key)
    log.debug("build new analysis record: %s", name)
    new_run = general_analysis('mip', version, analyzed_at)
    return {'case': name, 'run': new_run, 'samples': sample_ids}


def build_asset(asset_path, reference):
    """Build Asset record from reference data."""
    abs_path = Path(asset_path).abspath()
    new_asset = Asset(original_path=abs_path, category=reference['category'],
                      archive_type=reference.get('archive_type'))
    return new_asset



# -*- coding: utf-8 -*-
import logging
import pkg_resources

import click
from path import Path
import ruamel.yaml

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from housekeeper.cli.utils import run_orabort
from housekeeper import exc
from housekeeper.pipelines.mip.core import link_asset
from .mip import parse_mip
from .mip2 import parse_mip2
from .mip4 import parse_mip4
from .miparchive import parse as parse_miparchive
from .general import commit_analysis, check_existing
from .mip4.scout import prepare_scout

log = logging.getLogger(__name__)

LOADERS = {'mip': parse_mip, 'mip2': parse_mip2, 'mip4': parse_mip4,
           'mip38': parse_mip4, 'miparchive': parse_miparchive}


@click.command()
@click.option('-y', '--yes', is_flag=True, help='auto replace old runs')
@click.option('-f', '--force', is_flag=True, help='skip ALL validations')
@click.option('-r', '--references', type=click.File('r'))
@click.option('-p', '--pipeline', type=click.Choice(LOADERS.keys()),
              default='mip4')
@click.option('--replace', is_flag='replace identical runs')
@click.argument('config', type=click.File('r'))
@click.pass_context
def add(context, force, yes, replace, references, pipeline, config):
    """Add analyses from different pipelines."""
    manager = api.manager(context.obj['database'])
    config_data = ruamel.yaml.safe_load(config)
    if not references:
        default_ref = "pipelines/references/{}.yaml".format(pipeline)
        references = pkg_resources.resource_string("housekeeper", default_ref)
    reference_data = ruamel.yaml.safe_load(references)
    loader = LOADERS[pipeline]
    try:
        data = loader(config_data, reference_data, force=force)
    except exc.MalformattedPedigreeError as error:
        log.error("bad PED formatting: %s", error.message)
        context.abort()
    except exc.AnalysisNotFinishedError as error:
        log.error("analysis not finished: %s", error.message)
        context.abort()
    except exc.UnsupportedVersionError as error:
        new_or_old = 'old' if pipeline == 'mip' else 'new'
        log.error("pipeline too %s: %s", new_or_old, error.message)
        context.abort()

    records = link_records(data)
    case_name = records['case'].name
    old_run = check_existing(case_name, records['run'])
    if old_run:
        message = "identical run detected: {}".format(case_name)
        if force or replace:
            log.warn(message)
            api.delete(context.obj['root'], old_run)
            manager.commit()
        else:
            log.error(message)
            context.abort()

    try:
        commit_analysis(manager, context.obj['root'], records['case'], records['run'])
        click.echo("added new analysis: {}".format(case_name))
        sample_ids = ', '.join(sample.lims_id for sample in
                               records['run'].samples)
        click.echo("including samples: {}".format(sample_ids))
    except exc.AnalysisConflictError:
        click.echo("analysis output not removed: {}".format(case_name))


def link_records(data):
    """Fetch existing information."""
    case_obj = api.case(data['case'])
    if case_obj is None:
        raise ValueError("case not found: {}".format(data['case']))
    for sample_id in data['samples']:
        sample_obj = api.sample(sample_id)
        if sample_obj is None:
            raise ValueError("sample not found: {}".format(sample_id))
        data['run'].samples.append(sample_obj)

    for new_asset, sample_id in data['assets']:
        link_asset(data['run'], new_asset, sample=sample_id)
    return {'case': case_obj, 'run': data['run']}


@click.command()
@click.option('--no-link', is_flag=True, help='skip hard linking')
@click.option('-d', '--date', help="date of the particular run")
@click.option('-c', '--category', required=True)
@click.option('-s', '--sample', help='sample id to link assets to')
@click.option('-a', '--archive-type', type=click.Choice(['data', 'result']))
@click.argument('case_name')
@click.argument('asset_path')
@click.pass_context
def extend(context, no_link, date, category, sample, archive_type, case_name,
           asset_path):
    """Add an additional asset to a run."""
    manager = api.manager(context.obj['database'])
    run_obj = run_orabort(context, case_name, date)
    if sample:
        sample_map = {smpl.name: smpl for smpl in run_obj.samples}
        sample_obj = sample_map[sample]
    else:
        sample_obj = None
    new_asset = api.add_asset(run_obj, asset_path, category, archive_type,
                              sample=sample_obj)

    root_path = context.obj['root']
    run_root = get_rundir(root_path, run_obj.case.name, run_obj)
    filename = new_asset.basename()
    if no_link:
        new_path = asset_path
    else:
        new_path = run_root.joinpath(filename)
        log.debug("link asset: %s -> %s", new_asset.original_path, new_path)
        Path(new_asset.original_path).link(new_path)

    new_asset.path = new_path
    run_obj.assets.append(new_asset)
    log.info("add asset: %s", new_asset.path)
    manager.commit()


@click.command()
@click.option('-m', '--madeline-exe', type=click.Path(exists=True))
@click.option('-d', '--date', help="date of the particular run")
@click.option('-r', '--replace', is_flag=True)
@click.argument('case_name')
@click.pass_context
def scout(context, madeline_exe, date, replace, case_name):
    """Prepare a config and files for Scout."""
    madeline_exe = madeline_exe or context.obj['madeline_exe']
    manager = api.manager(context.obj['database'])
    root_path = context.obj['root']
    run_obj = run_orabort(context, case_name, date)
    if not run_obj.pipeline_version.startswith('v4'):
        log.error("unsupported pipeline version: %s", run_obj.pipeline_version)
        context.abort()

    existing_conf = (api.assets(category='scout-config', run_id=run_obj.id)
                        .first())
    if existing_conf:
        if replace:
            log.info("deleting existing scout config: %s", existing_conf.path)
            api.delete_asset(existing_conf)
            existing_mad = (api.assets(category='madeline', run_id=run_obj.id)
                               .first())
            if existing_mad:
                log.info("deleting existing madeline: %s", existing_mad.path)
                api.delete_asset(existing_mad)
            manager.commit()
        else:
            log.error("scout config already generated")
            context.abort()

    prepare_scout(run_obj, root_path, madeline_exe)
    manager.commit()



# -*- coding: utf-8 -*-
"""Flask app module."""
from __future__ import division
import logging
import os

from dateutil.parser import parse as parse_date
from flask import abort, flash, Flask, render_template, request, redirect
from flask_alchy import Alchy
from flask_bootstrap import Bootstrap
from flask_login import current_user, login_required
from werkzeug.contrib.fixers import ProxyFix

from housekeeper.constants import EXTRA_STATUSES
from housekeeper.store import Model, api, Sample, Case, AnalysisRun
from housekeeper.store.models import User
from .admin import UserManagement


log = logging.getLogger(__name__)

app = Flask(__name__)
app.wsgi_app = ProxyFix(app.wsgi_app)

# configuration
SECRET_KEY = os.environ.get('SECRET_KEY') or 'thisIsNotSecret!'
SQLALCHEMY_DATABASE_URI = os.environ['SQLALCHEMY_DATABASE_URI']
GOOGLE_OAUTH_CLIENT_ID = os.environ['GOOGLE_OAUTH_CLIENT_ID']
GOOGLE_OAUTH_CLIENT_SECRET = os.environ['GOOGLE_OAUTH_CLIENT_SECRET']
# serve boostrap assets locally is running in debug mode
BOOTSTRAP_SERVE_LOCAL = 'FLASK_DEBUG' in os.environ
if 'mysql' in SQLALCHEMY_DATABASE_URI:  # pragma: no cover
    SQLALCHEMY_POOL_RECYCLE = 3600
SQLALCHEMY_TRACK_MODIFICATIONS = False
TEMPLATES_AUTO_RELOAD = True
app.config.from_object(__name__)

# database setup
db = Alchy(Model=Model)
user = UserManagement(db, User)

app.jinja_env.globals['EXTRA_STATUSES'] = EXTRA_STATUSES


@app.route('/')
def index():
    """Show dashboard."""
    if not current_user.is_authenticated:
        return render_template('index.html')

    plots = []
    for category in ['wgs', 'wes', 'tga']:
        all_samples = Sample.query.filter_by(category=category)
        samples_count = all_samples.count()
        sequenced_count = all_samples.filter(Sample.sequenced_at != None).count()
        plots.append({
            'category': category,
            'total': samples_count,
            'sequenced': sequenced_count,
            'sequenced_percent': (sequenced_count / samples_count) * 100,
            'to_sequenced': api.to_sequenced(db.session, category),
            'to_analyzed': api.to_analyzed(db.session, category),
        })

    all_cases = Case.query
    cases_count = all_cases.count()
    analyzed_cases = (all_cases.join(Case.runs)
                               .filter(AnalysisRun.analyzed_at != None)
                               .count())
    data = {
        'cases': cases_count,
        'analyzed_percent': (analyzed_cases / cases_count) * 100,
        'plots': plots,
    }
    return render_template('dashboard.html', **data)


@app.route('/cases')
@login_required
def cases():
    """Overview all loaded cases."""
    missing_category = request.args.get('missing')
    if missing_category == 'empty':
        missing_category = None
    page = int(request.args.get('page', '1'))
    qargs = {
        'query_str': request.args.get('query_str'),
        'per_page': 30,
        'missing': missing_category,
    }
    version = 'v4' if qargs['missing'] == 'delivered' else None
    cases_q = api.cases(query_str=qargs['query_str'], missing=qargs['missing'],
                        version=version)
    cases_page = cases_q.paginate(page, per_page=qargs['per_page'])
    return render_template('cases.html', cases=cases_page, qargs=qargs)


@app.route('/cases/<name>')
@login_required
def case(name):
    """Show more information about a case."""
    case_obj = api.case(name)
    if case_obj is None:
        return abort(404)
    return render_template('case.html', case=case_obj)


@app.route('/cases/<case_name>/postpone', methods=['POST'])
@login_required
def case_postpone(case_name):
    """Postpone clean up date for the latest run of a case."""
    run_obj = api.runs(case_name).first()
    if run_obj is None:
        return abort(404)
    api.postpone(run_obj)
    db.commit()
    return redirect(request.referrer)


@app.route('/cases/<case_name>/onhold', methods=['POST'])
@login_required
def case_onhold(case_name):
    """Toggle onhold status for a case."""
    case_obj = api.case(case_name)
    if case_obj is None:
        return abort(404)
    case_obj.is_onhold = False if case_obj.is_onhold else True
    db.commit()
    return redirect(request.referrer)


@app.route('/samples')
@login_required
def samples():
    """Overview all loaded samples."""
    page = int(request.args.get('page', '1'))
    status_to = request.args.get('status_to')
    qargs = {
        'query_str': request.args.get('query_str'),
        'per_page': 30,
        'status_to': status_to
    }
    samples_q = api.samples(query_str=qargs['query_str'], status_to=status_to)
    samples_page = samples_q.paginate(page, per_page=qargs['per_page'])
    return render_template('samples.html', samples=samples_page, qargs=qargs)


@app.route('/comments/<int:case_id>', methods=['POST'])
@login_required
def comments(case_id):
    """Interact with comments."""
    case_obj = Case.query.get(case_id)
    if case_obj is None:
        return abort(404)
    new_comment = request.form['comment']
    case_obj.comment = new_comment
    db.commit()
    return redirect(request.referrer)


@app.route('/queue')
@login_required
def queue():
    """Overview of status of samples/cases."""
    data = dict(
        samples_to_receive=api.samples(status_to='receive'),
        samples_to_sequence=api.samples(status_to='sequence'),
        cases_to_analyze=api.cases(missing='analyzed', onhold=False),
        cases_to_deliver=api.cases(missing='delivered'),
    )
    return render_template('queue.html', **data)


@app.route('/cases/<int:case_id>/manual', methods=['POST'])
@login_required
def case_manual(case_id):
    """Mark a case as manual."""
    case_obj = Case.query.get(case_id)
    if case_obj is None:
        return abort(404)
    case_obj.is_manual = True
    db.commit()
    return redirect(request.referrer)


@app.route('/runs/<int:run_id>/delivered', methods=['POST'])
@login_required
def run_delivered(run_id):
    """Mark an analysis run of a case as delivered."""
    run_obj = AnalysisRun.query.get(run_id)
    if run_obj is None:
        flash("run not found: %s", run_id)
    else:
        run_obj.delivered_at = parse_date(request.form['delivery_date'])
        db.commit()
    return redirect(request.referrer)


@app.route('/samples/<int:sample_id>/sequenced', methods=['POST'])
@login_required
def sample_sequenced(sample_id):
    """Mark a sample as sequenced."""
    sample_obj = Sample.query.get(sample_id)
    if sample_obj is None:
        return abort(404)
    sequence_date = parse_date(request.form['when'])
    sample_obj.sequenced_at = sequence_date
    db.commit()
    return redirect(request.referrer)


# hookup extensions to app
Bootstrap(app)
db.init_app(app)
user.init_app(app)



# -*- coding: utf-8 -*-
import logging

from path import Path

from housekeeper.exc import MissingFileError
from .build import build_analysis, build_asset
from .parse import prepare_inputs, parse_references
from .prepare import prepare_run

log = logging.getLogger(__name__)


def build_assets(new_references, force=False):
    """Build assets from references."""
    for ref_data in new_references:
        log.debug("working on: %s", ref_data['reference']['category'])
        if not Path(ref_data['path']).exists():
            is_required = ref_data['reference'].get('required', True)
            if force or not is_required:
                log.warn("skipping asset: %s", ref_data['path'])
                continue
            else:
                raise MissingFileError(ref_data['path'])

        new_asset = build_asset(ref_data['path'], ref_data['reference'])
        yield new_asset, ref_data.get('sample')


def link_asset(run_obj, new_asset, sample=None):
    if sample:
        run_obj.sample_map[sample].assets.append(new_asset)
    run_obj.assets.append(new_asset)


def parse_mip(config_data, reference_data, force=False):
    """Parse MIP analysis output."""
    # 1. parse segments of input data
    segments = prepare_inputs(config_data)
    # 2. post-process the output a bit
    prepare_run(segments, force=force)
    # 3. build the records
    new_data = build_analysis(segments)
    # 4. parse references
    new_refs = parse_references(reference_data, segments)
    # 5. build assets from references + link to new records
    new_assets = build_assets(new_refs, force=force)
    return {
        'case': new_data['case'],
        'samples': new_data['samples'],
        'run': new_data['run'],
        'assets': new_assets,
    }



# -*- coding: utf-8 -*-
import logging

import click
import ruamel.yaml

from housekeeper.store import api

log = logging.getLogger(__name__)


@click.command('mip-sex')
@click.option('-s', '--sample', help='limit to a single sample')
@click.argument('case_id', required=False)
@click.pass_context
def mip_sex(context, sample, case_id):
    """Parse out analysis determined sex of sample."""
    if sample and case_id is None:
        sample_obj = api.sample(sample)
        if sample_obj is None:
            log.warn('sorry, sample not found')
            context.abort()
        case_id = sample_obj.case_id

    qc_metrics = api.assets(case_id, category='qc').first()
    qcm_data = ruamel.yaml.safe_load(qc_metrics.path)
    sample_sexes = parse_mipsex(qcm_data)
    if sample:
        click.echo(sample_sexes[sample])
    else:
        for sample_id, sex in sample_sexes.items():
            click.echo("{}: {}".format(sample_id, sex))


def parse_mipsex(qcm_data):
    """Parse out sample sex from qc metrics data."""
    samples_sex = parse_samples(qcm_data)
    sexes = {sample_id: sex for sample_id, sex in samples_sex}
    return sexes


def parse_samples(qcm_data):
    """Parse out the relevant sample information."""
    fam_key = qcm_data.keys()[0]
    for segment_id, values in qcm_data[fam_key].items():
        if segment_id != fam_key:
            # sample data entry, find main section
            for sebsection_id, data in values.items():
                if '_lanes_' in sebsection_id:
                    yield segment_id, data['ChanjoSexCheck']['gender']



# -*- coding: utf-8 -*-
"""Build Scout config from MIP output."""
import logging
from functools import partial

from path import Path
import ruamel.yaml

from housekeeper.store import api
from housekeeper.store.utils import get_rundir
from . import madeline

SEX_MAP = {'1': 'male', '2': 'female', '0': 'unknown'}
PHENOTYPE_MAP = {'1': 'unaffected', '2': 'affected'}

log = logging.getLogger(__name__)


def prepare_scout(run_obj, root_path, madeline_exe):
    """Prepare files for Scout upload."""
    run_root = Path(get_rundir(root_path, run_obj.case.name, run=run_obj))
    config_data = build_config(run_obj)

    pedigree = api.assets(category='pedigree', run_id=run_obj.id).one()
    # extract external sample ids
    ped_yaml = api.assets(category='pedigree-yaml', run_id=run_obj.id).one()
    with Path(ped_yaml.path).open() as in_handle:
        ped_data = ruamel.yaml.safe_load(in_handle)
    sample_map = {sample['sample_id']: sample['sample_name']
                  for sample in ped_data['samples']}
    madeline_path = build_madeline(pedigree.path, run_root, madeline_exe,
                                   sample_map)
    if madeline_path:
        mad_asset = api.add_asset(run_obj, madeline_path, 'madeline')
        mad_asset.path = str(madeline_path)
        log.info("add madeline asset: %s", mad_asset.path)
        run_obj.assets.append(mad_asset)
        config_data['madeline'] = mad_asset.path

    # save the new scout config
    config_path = Path(run_root).joinpath('scout.conf.yaml')
    with open(config_path, 'w') as out_handle:
        ruamel.yaml.dump(config_data, stream=out_handle, Dumper=ruamel.yaml.RoundTripDumper)

    scout_asset = api.add_asset(run_obj, config_path, 'scout-config')
    scout_asset.path = config_path
    log.info("add scout config asset: %s", scout_asset.path)
    run_obj.assets.append(scout_asset)


def build_madeline(pedigree, run_root, madeline_exe, sample_map):
    """Build and add madeline file to the database."""
    try:
        with open(pedigree, 'r') as in_handle:
            svg_content = madeline.run(in_handle, sample_map, exe=madeline_exe)
        madeline_path = Path(run_root).joinpath('madeline.xml')
        with open(madeline_path, 'w') as out_handle:
            out_handle.write(svg_content)
        return madeline_path
    except (madeline.SinglePedigreeError,
            madeline.MadelineIncompatibleError) as error:
        log.warn("can't visualize pedigree: %s", error.message)
        return None


def build_config(run_obj):
    """Build a Scout config from a MIP run."""
    customer, family = run_obj.case.name.split('-', 1)
    run_asset = partial(api.assets, run_id=run_obj.id)
    vcf = run_asset(category='vcf-clinical-bin').one()
    vcf_sv = run_asset(category='vcf-clinical-sv-bin').first()
    vcf_research = run_asset(category='vcf-research-bin').one()
    vcf_research_sv = run_asset(category='vcf-research-sv-bin').first()
    sampleinfo = run_asset(category='sampleinfo').one()

    # start from pedigree YAML and add additional information
    pedigree_yaml = run_asset(category='pedigree-yaml').one()
    with open(pedigree_yaml.path, 'r') as in_handle:
        data = ruamel.yaml.safe_load(in_handle)

    with open(sampleinfo.path, 'r') as in_handle:
        si_data = ruamel.yaml.safe_load(in_handle)
    rank_model = si_data['program']['rankvariant']['rank_model']['version']
    genome_build = (si_data['human_genome_build']['source'] +
                    si_data['human_genome_build']['version'])

    data['vcf_snv'] = vcf.path
    data['vcf_sv'] = vcf_sv.path if vcf_sv else None
    data['rank_model_version'] = float(rank_model)
    data['analysis_date'] = run_obj.analyzed_at
    data['human_genome_build'] = genome_build
    data['vcf_snv_research'] = vcf_research.path
    data['vcf_sv_research'] = vcf_research_sv.path if vcf_research_sv else None

    for ped_sample in data['samples']:
        lims_id = ped_sample['sample_id']
        bam_file = run_asset(category='bam', sample=lims_id).one()
        ped_sample['bam_path'] = bam_file.path
        if ped_sample['mother'] == 0:
            del ped_sample['mother']
        if ped_sample['father'] == 0:
            del ped_sample['father']

    return data



# -*- coding: utf-8 -*-
import logging

from housekeeper.pipelines.general.add import analysis as general_analysis
from housekeeper.pipelines.mip.parse import parse_references
from housekeeper.pipelines.mip.core import build_assets
from .parse import prepare_inputs
from .prepare import prepare_run

log = logging.getLogger(__name__)


def parse_mip4(config_data, reference_data, force=False):
    """Parse MIP analysis output."""
    # 1. parse segments of input data
    segments = prepare_inputs(config_data)
    # 2. post-process the output a bit
    prepare_run(segments, force=force)
    # 3. build the records
    new_data = build_analysis(segments)
    # 4. parse references
    new_refs = parse_references(reference_data, segments=segments)
    # 5. build assets from references + link to new records
    new_assets = build_assets(new_refs, force=force)
    return {
        'case': new_data['case'],
        'samples': new_data['samples'],
        'run': new_data['run'],
        'assets': new_assets,
    }


def build_analysis(segments):
    """Prepare info for a MIP analysis."""
    version = segments['family']['mip_version']
    analyzed_at = segments['family']['analysis_date']
    sample_ids = segments['config']['sample_ids']
    customer = segments['pedigree']['owner']
    genome_build = segments['family']['human_genome_build']
    reference_genome = ''.join([genome_build['source'], genome_build['version']])
    name = "{}-{}".format(customer, segments['config']['family_id'])
    log.debug("build new analysis record: %s", name)
    new_run = general_analysis('mip', version, analyzed_at,
                               reference_genome=reference_genome)
    return {'case': name, 'run': new_run, 'samples': sample_ids}
